import math
import numdifftools as nd


class taylor:

	def __init__(self, funcx=math.log, degree=5, fpoint=0):
		self.degree = degree
		self.fpoint = fpoint
		self.coeffs = []

		for i in range(0, self.degree+1):
			ider = nd.Derivative(funcx, n=i)
			self.coeffs.append(ider(self.fpoint)/math.factorial(i))

	def print_equation(self):
		eqn_string = ""
		for i in range(self.degree + 1):
			if self.coeffs[i] != 0:
				eqn_string += str(self.coeffs[i]) + ("(x-{})^{}".format(self.fpoint, i) if i > 0 else "") + " + "
		eqn_string = eqn_string[:-3] if eqn_string.endswith(" + ") else eqn_string
		print(eqn_string)

	def get_coeffs(self):
		return self.coeffs

	def print_coeffs(self):
		print(self.coeffs)

	def predict(self, x):
		fx = 0
		for i in range(len(self.coeffs)):
			fx += self.coeffs[i] * ((x - self.fpoint)**i)  # coefficient * nth term
		return fx

	def approximate_derivative(self, x):
		value = 0
		for i in range(1, len(self.coeffs)): # skip the first value (constant) as the derivative is 0
			value += self.coeffs[i] * i * ((x - self.fpoint)**(i-1)) # differentiate each term: x^n => n*x^(n-1)
		return value

	def approximate_integral(self, x0, x1):
		value = 0
		for i in range(len(self.coeffs)):
			value += ((self.coeffs[i] * (1/(i+1)) * ((x1 - self.fpoint)**(i+1))) -
			(self.coeffs[i] * (1/(i+1)) * ((x0 - self.fpoint)**(i+1)))) # integrate each term: x^n => (1/n+1)*x^(n+1)
		return value



import numpy as np
from tensorflow import keras


class tf_ann:

	def __init__(self, degree=5, optimizer='adam', loss=keras.losses.MeanSquaredError()):
		self.degree = degree
		self.model = keras.Sequential([
			keras.layers.Dense(1, input_dim=self.degree, activation=keras.activations.linear),
			keras.layers.Dense(1, activation=keras.activations.linear)
			])
		self.model.compile(optimizer=optimizer, loss=loss)
		self.model.summary()


	def fit(self, funcx=np.log, frange=[-1,1], points=2**11, batch_size=32, epochs=512):
		self.ary = np.linspace(frange[0], frange[1], points)
		x_train = []
		y_train = []

		for i in self.ary:
			xal = []
			for d in range(self.degree):
				xal.append(i**(d+1))
			x_train.append(xal)
			y_train.append(funcx(i))

		history = self.model.fit(x_train, y_train, batch_size=batch_size, epochs=epochs)

		w0, b0 = self.model.layers[0].get_weights()
		w1, b1 = self.model.layers[1].get_weights()
		wts = w0 * w1
		bias = (b0*w1) + b1

		self.coeffs = []
		self.coeffs.append(bias[0][0])
		for w in wts:
			self.coeffs.append(w[0])

		return self.coeffs


	def predict(self, x):
		result = 0
		for i in range(len(self.coeffs)):
			result += self.coeffs[i] * (x**i)
		return result



import unittest
from chiku import chebyshev
import numpy as np
import random

class Test_Chebyshev(unittest.TestCase):
	def test_chebyshev(self):
		def sigmoid(x):
			return 1/(1+np.exp(-x))

		self.poly = chebyshev.chebyshev(sigmoid)
		self.rnum = random.random()
		self.assertAlmostEqual(self.poly.predict(self.rnum), sigmoid(self.rnum), 2)

if __name__ == '__main__':
	unittest.main()



import unittest
from chiku import tf_ann
import numpy as np
import random

class Test_TF_ANN(unittest.TestCase):
	def test_tf_ann(self):
		def sigmoid(x):
			return 1/(1+np.exp(-x))

		self.poly = tf_ann.tf_ann()
		self.poly.fit(sigmoid, epochs=256)

		self.rnum = random.random()
		self.assertAlmostEqual(self.poly.predict(self.rnum), sigmoid(self.rnum), 2)

if __name__ == '__main__':
	unittest.main()
